#include <io.h>
#include <stdio.h>
#include <direct.h>
#include <windows.h>
#include <sys/stat.h>
#include <errno.h>

#define APP_NAME    "dubl"

#ifndef _MAX_PATH
#define _MAX_PATH   260
#endif

// ------------------------------------

void usage ()
{
    fprintf ( stderr,
        "Usage: " APP_NAME " <source dir> <target dir> [command + parameters...]\n"
        "\n"
        "Dublicates directory structure from <source dir> to <target dir>.\n"
        "Optional parameter [command] will be applied in all directories under source.\n"
        "\n"
        "There are three variables that can be used with parameters to command:\n"
        "#1 = name of source file without path\n"
        "#2 = name of target directory\n"
        "#3 = like '#1' but without extension\n"
        "\n"
        "Example of copying all files from source to target:\n"
        "  " APP_NAME " copy #1 #2\\#1\n"
        "\n"
    );
}

int make_command ( char *buffer, const char *command, const char *filename, const char *dstpath )
{
    char        *d = buffer;
    const char  *s = command;
    int         quotes = 0;

    while ( *s ) {
        char ch = *s++;

        if ( ch != '#' ) {
            if ( ch == ' ' && quotes ) {
                *d++ = '"';
                quotes = 0;
            }
            *d++ = ch;
        } else {
            ch = *s++;

            if ( ch == '1' || ch == '2' || ch == '3' ) {
                if ( !quotes ) {
                    *d++ = '"';
                    quotes = 1;
                }
                if ( ch == '1' ) {
                    d += sprintf ( d, "%s", filename );
                } else if ( ch == '2' ) {
                    d += sprintf ( d, "%s", dstpath );
                } else {
                    char temp[_MAX_FNAME];
                    _splitpath ( filename, NULL, NULL, temp, NULL );
                    d += sprintf ( d, "%s", temp );
                }
            } else {
                *d++ = '#';
                if ( ch == '\0' ) break;
                *d++ = ch;
            }
        }
    }

    if ( quotes ) *d++ = '"';
    *d = '\0';

    return 0;
}

unsigned int scan_files ( const char *command, const char *dstpath )
{
    char                cmd[16384];
    struct _finddata_t  fd;
    long                hFile;
    unsigned int        errors = 0;

    if ( (hFile = _findfirst ("*.*", &fd)) != -1L ) {
        if ( !(fd.attrib & _A_SUBDIR) && !(fd.attrib & _A_HIDDEN) ) {
            make_command ( cmd, command, fd.name, dstpath );
            system ( cmd );
        }

        while ( _findnext (hFile, &fd) == 0 ) {
            if ( !(fd.attrib & _A_SUBDIR) && !(fd.attrib & _A_HIDDEN) ) {
                make_command ( cmd, command, fd.name, dstpath );
                system ( cmd );
            }
        }

        _findclose ( hFile );
    }

    return errors;
}

unsigned int scan_path ( const char *srcpath, const char *dstpath, const char *command )
{
    char                cwdpath[_MAX_PATH];
    char                temp[_MAX_PATH*2];
    struct _finddata_t  fd;
    long                hFile;
    unsigned int        errors = 0;

    _getcwd ( cwdpath, _MAX_PATH );
    if ( srcpath ) _chdir ( srcpath );
    if ( command ) scan_files ( command, dstpath );

    if ( (hFile = _findfirst ("*.*", &fd)) != -1L ) {
        if ( (fd.attrib & _A_SUBDIR) && !(fd.attrib & _A_HIDDEN) ) {
            if ( strcmp (fd.name, ".") && strcmp (fd.name, "..") ) {
                sprintf ( temp, "%s\\%s", srcpath, fd.name );
                if ( stricmp (temp, dstpath) ) {
                    sprintf ( temp, "%s\\%s", dstpath, fd.name );
                    if ( _mkdir (temp) != 0 && errno != EEXIST ) {
                        fprintf ( stderr, APP_NAME ": Failed to create directory '%s'\n", temp );
                        errors++;
                    } else {
                        errors += scan_path ( fd.name, temp, command );
                    }
                }
            }
        }

        while ( _findnext (hFile, &fd) == 0 ) {
            if ( (fd.attrib & _A_SUBDIR) && !(fd.attrib & _A_HIDDEN) ) {
                if ( strcmp (fd.name, ".") && strcmp (fd.name, "..") ) {
                    sprintf ( temp, "%s\\%s", srcpath, fd.name );
                    if ( stricmp (temp, dstpath) ) {
                        sprintf ( temp, "%s\\%s", dstpath, fd.name );
                        if ( _mkdir (temp) != 0 && errno != EEXIST ) {
                            fprintf ( stderr, APP_NAME ": Failed to create directory '%s'\n", temp );
                            errors++;
                        } else {
                            errors += scan_path ( fd.name, temp, command );
                        }
                    }
                }
            }
        }

        _findclose ( hFile );
    }

    _chdir ( cwdpath );

    return errors;
}

int isdir ( const char *path )
{
    struct _stat st;
    if ( _stat (path, &st) != 0 ) return 0;
    return st.st_mode & _S_IFDIR;
}

int main ( int argc, char **argv )
{
    char *command;
    char srcpath[_MAX_PATH];
    char dstpath[_MAX_PATH];
    char *p;
    unsigned int errors;

    if ( argc < 3 ) {
        usage ();
        return 1;
    }

    if ( !isdir (argv[1]) ) {
        fprintf ( stderr, "'%s' is not a directory\n", argv[1] );
        return 1;
    }
    if ( !isdir (argv[2]) ) {
        fprintf ( stderr, "'%s' is not a directory\n", argv[2] );
        return 1;
    }

    GetFullPathName ( argv[1], _MAX_PATH, srcpath, &p );
    GetFullPathName ( argv[2], _MAX_PATH, dstpath, &p );
    GetLongPathName ( srcpath, srcpath, _MAX_PATH );
    GetLongPathName ( dstpath, dstpath, _MAX_PATH );

    if ( !stricmp (srcpath, dstpath) ) {
        fprintf ( stderr, "Source and destination directories are the same.\n" );
        return 1;
    }

    if ( argc > 3 ) {
        int i, len = 0;
        char *p;

        for ( i = 3; i < argc; i++ ) {
            len += strlen ( argv[i] ) + 1;
        }

        if ( (command = (char *)malloc (len)) == NULL ) {
            fprintf ( stderr, APP_NAME ": Memory allocation failed.\n" );
            return 1;
        }

        p = command;
        p += sprintf ( p, "%s", argv[3] );

        for ( i = 4; i < argc; i++ ) {
            p += sprintf ( p, " %s", argv[i] );
        }
    } else {
        command = NULL;
    }

    errors = scan_path ( srcpath, dstpath, command );
    free ( command );

    return (int)errors;
}
